LANGUAGE = {
    "adjective": [
        "third",
        "obsessed",
        "seventh",
        "silent",
        "blue",
        "purple",
        "sacred",
        "hot",
        "lovely",
        "captured",
        "trembling",
        "burning",
        "professional",
        "first",
        "luscious",
        "black",
        "diamond",
        "misty",
        "next",
        "willing",
        "all",
        "lonely",
        "swollen",
        "forgotten",
        "no",
        "elemental",
        "what",
        "silver",
        "red",
        "living",
        "last",
        "sleeping",
        "bloody",
        "lost",
        "invisible",
        "whispering",
        "dark",
        "white",
        "naked",
        "which",
        "bare",
        "hidden",
        "fallen",
        "dangerous",
        "sucking",
        "wild",
        "ragged",
        "licking",
        "devoted",
        "kissing",
        "grey",
        "prized",
        "green",
        "missing",
        "silky",
        "growing",
        "darkest",
        "wet",
        "rough",
        "cracked",
        "bold",
        "bound",
        "slithering",
        "unwilling",
        "vacant",
        "delicious",
        "dying",
        "only",
        "erect",
        "some",
        "smooth",
        "absent",
        "eager",
        "playful",
        "silken",
        "falling",
        "laughing",
        "broken",
        "entwined",
        "rising",
        "hard",
        "sharp",
        "dwindling",
        "each",
        "splintered",
        "silvery",
        "stolen",
        "wanton",
        "final",
        "twinkling",
        "cold",
        "weeping",
        "stripped",
        "magnificent",
        "ravaged",
        "deep",
        "frozen",
        "shadowy",
        "emerald",
        "azure",
        "every",
    ],
    "noun": [
        "dream",
        "dreamer",
        "dreams",
        "waves",
        "sword",
        "kiss",
        "sex",
        "lover",
        "slave",
        "slaves",
        "pleasure",
        "servant",
        "servants",
        "snake",
        "soul",
        "touch",
        "men",
        "women",
        "gift",
        "scent",
        "ice",
        "snow",
        "night",
        "silk",
        "secret",
        "secrets",
        "game",
        "fire",
        "flame",
        "flames",
        "husband",
        "wife",
        "man",
        "woman",
        "boy",
        "girl",
        "truth",
        "edge",
        "boyfriend",
        "girlfriend",
        "body",
        "captive",
        "male",
        "wave",
        "predator",
        "female",
        "healer",
        "trainer",
        "teacher",
        "hunter",
        "obsession",
        "hustler",
        "consort",
        "dream",
        "dreamer",
        "dreams",
        "rainbow",
        "dreaming",
        "flight",
        "flying",
        "soaring",
        "wings",
        "mist",
        "sky",
        "wind",
        "winter",
        "misty",
        "river",
        "door",
        "gate",
        "cloud",
        "fairy",
        "dragon",
        "end",
        "blade",
        "beginning",
        "tale",
        "tales",
        "emperor",
        "prince",
        "princess",
        "willow",
        "birch",
        "petals",
        "destiny",
        "theft",
        "thief",
        "legend",
        "prophecy",
        "spark",
        "sparks",
        "stream",
        "streams",
        "waves",
        "sword",
        "darkness",
        "swords",
        "silence",
        "kiss",
        "butterfly",
        "shadow",
        "ring",
        "rings",
        "emerald",
        "storm",
        "storms",
        "mists",
        "world",
        "worlds",
        "alien",
        "lord",
        "lords",
        "ship",
        "ships",
        "star",
        "stars",
        "force",
        "visions",
        "vision",
        "magic",
        "wizards",
        "wizard",
        "heart",
        "heat",
        "twins",
        "twilight",
        "moon",
        "moons",
        "planet",
        "shores",
        "pirates",
        "courage",
        "time",
        "academy",
        "school",
        "rose",
        "roses",
        "stone",
        "stones",
        "sorcerer",
        "shard",
        "shards",
        "slave",
        "slaves",
        "servant",
        "servants",
        "serpent",
        "serpents",
        "snake",
        "soul",
        "souls",
        "savior",
        "spirit",
        "spirits",
        "voyage",
        "voyages",
        "voyager",
        "voyagers",
        "return",
        "legacy",
        "birth",
        "healer",
        "healing",
        "year",
        "years",
        "death",
        "dying",
        "luck",
        "elves",
        "tears",
        "touch",
        "son",
        "sons",
        "child",
        "children",
        "illusion",
        "sliver",
        "destruction",
        "crying",
        "weeping",
        "gift",
        "word",
        "words",
        "thought",
        "thoughts",
        "scent",
        "ice",
        "snow",
        "night",
        "silk",
        "guardian",
        "angel",
        "angels",
        "secret",
        "secrets",
        "search",
        "eye",
        "eyes",
        "danger",
        "game",
        "fire",
        "flame",
        "flames",
        "bride",
        "husband",
        "wife",
        "time",
        "flower",
        "flowers",
        "light",
        "lights",
        "door",
        "doors",
        "window",
        "windows",
        "bridge",
        "bridges",
        "ashes",
        "memory",
        "thorn",
        "thorns",
        "name",
        "names",
        "future",
        "past",
        "history",
        "something",
        "nothing",
        "someone",
        "nobody",
        "person",
        "man",
        "woman",
        "boy",
        "girl",
        "way",
        "mage",
        "witch",
        "witches",
        "lover",
        "tower",
        "valley",
        "abyss",
        "hunter",
        "truth",
        "edge",
    ],
}

LABELS = {
    "storage.googleapis.com": "Google API Storage",
    "v.vrv.co": "VRV",
    "pl.crunchyroll.com": "Crunchyroll",
}



import regex
import yarl


CONTENT_ID_REGEX = regex.compile(r"embed-6/([^?#&/.]+)")


def extract(session, url, **opts):

    content_id = CONTENT_ID_REGEX.search(url).group(1)
    sources = session.get(
        "https://{}/ajax/embed-6/getSources".format(yarl.URL(url).host),
        params={"id": content_id},
    ).json()

    subtitles = [
        _.get("file") for _ in sources.get("tracks") if _.get("kind") == "captions"
    ]

    def yielder():
        for _ in sources.get("sources"):
            yield {"stream_url": _.get("file"), "subtitle": subtitles}

        for _ in sources.get("sourcesBackup"):
            yield {"stream_url": _.get("file"), "subtitle": subtitles}

    return list(yielder())



import base64

import regex
import yarl
from Cryptodome.Cipher import AES

GOGOANIME_SECRET = b"257465385929383" b"96764662879833288"


def get_quality(url_text):
    match = regex.search(r"(\d+) P", url_text)

    if not match:
        return None

    return int(match.group(1))


def pad(data):
    length = 16 - (len(data) % 16)
    return data + chr(length) * length


def extract(session, url, **opts):
    """
    Extract content off of GogoAnime.
    """
    parsed_url = yarl.URL(url)
    next_host = "https://{}/".format(parsed_url.host)

    encrypted_ajax = base64.b64encode(
        AES.new(GOGOANIME_SECRET, AES.MODE_CBC, iv=b"4206913378008135").encrypt(
            pad(parsed_url.query.get("id").replace("%3D", "=")).encode()
        )
    )

    content = session.get(
        "{}encrypt-ajax.php".format(next_host),
        params={"id": encrypted_ajax.decode(), "time": "69420691337800813569"},
        headers={"x-requested-with": "XMLHttpRequest"},
    ).json()

    def yielder():
        for origin in content.get("source"):
            yield {
                "stream_url": origin.get("file"),
                "quality": get_quality(origin.get("label", "")),
                "headers": {"referer": next_host},
            }

        for backups in content.get("source_bk"):
            yield {
                "stream_url": backups.get("file"),
                "quality": get_quality(origin.get("label", "")),
                "headers": {"referer": next_host},
            }

    return list(yielder())



from ....config import NINEANIME
from ...helper import construct_site_based_regex
from .. import crunchyroll, gogoanime, twistmoe, zoro

REGEX = construct_site_based_regex(
    NINEANIME, extra_regex=r"/watch/[^&?/]+\.(?P<slug>[^&?/]+)"
)

ALTERNATIVES = "https://animixplay.to/assets/rec/{}.json"
SLUG_SEARCH = "https://raw.githubusercontent.com/MALSync/MAL-Sync-Backup/master/data/pages/9anime/{}.json"


def safe_iter(session, check, provider, url):
    match = provider.REGEX.search(url)

    if not match:
        return None, None

    genexp = provider.fetcher(session, url, check, match)

    try:
        return next(genexp), genexp
    except StopIteration:
        return None, None


modules = {
    "Crunchyroll": crunchyroll,
    "Zoro": zoro,
    "Twistmoe": twistmoe,
    "Gogoanime": gogoanime,
}

PRIORITY = {"Crunchyroll": 0, "Zoro": 1, "Twistmoe": 2, "Gogoanime": 3}


def fetcher(session, url, check, match):

    mal_sync = session.get(SLUG_SEARCH.format(match.group("slug")))

    if mal_sync.status_code == 404:
        return

    parsed = mal_sync.json()
    alts = session.get(ALTERNATIVES.format(parsed.get("malId"))).json()

    for site, contents in sorted(
        alts.items(), key=lambda x: PRIORITY.get(x[0], float("inf"))
    ):

        if site not in modules:
            continue

        module = modules.get(site)

        for content in contents:
            _, genexp = safe_iter(session, check, module, content.get("url"))
            if genexp is None:
                continue
            else:
                yield _
                yield from genexp
                return



import regex

from functools import partial

import lxml.html as htmlparser

from ....config import TENSHI
from ...helper import construct_site_based_regex, uwu

REGEX = construct_site_based_regex(TENSHI, extra_regex=r"/anime/([^?&/]+)")


def extract_urls(session, episode_page):
    episode_page_content = session.get(episode_page)
    embed_page = (
        htmlparser.fromstring(episode_page_content.text).cssselect("iframe") or [{}]
    )[0].get("src")
    streams_page = session.get(embed_page, headers={"referer": episode_page})
    cookies = "__ddg1={}; __ddg2={}".format(
        session.cookies.get("__ddg1", domain=".tenshi.moe"),
        session.cookies.get("__ddg2", domain=".tenshi.moe"),
    )
    yield from (
        {
            "quality": int(_.group(2)),
            "stream_url": _.group(1),
            "headers": {"cookie": cookies},
        }
        for _ in regex.finditer(
            r"src: '(.+?)'.+?size: (\d+)", streams_page.text, flags=regex.S
        )
    )


def fetcher(session, url, check, match):
    uwu.bypass_ddos_guard(session, TENSHI)
    url = match.group(0)

    episode_list_page = session.get(url)
    count = int(
        htmlparser.fromstring(episode_list_page.text)
        .cssselect("span.badge")[0]
        .text_content()
    )

    for episode in range(1, count + 1):
        if check(episode):
            yield partial(
                lambda c: [*extract_urls(session, c)], "{}/{:d}".format(url, episode)
            ), episode



import regex

EMBED_URL_REGEX = regex.compile(r"(.+?/)(?:embed|e)/([^?/#&])")


def extract(session, url, **opts):
    host, slug = EMBED_URL_REGEX.search(url).group(1, 2)

    vidstream_info = session.get(
        "{}info/{}".format(host, slug), headers={"referer": host}
    )
    return [
        {"stream_url": content.get("file", ""), "headers": {"referer": url}}
        for content in vidstream_info.json().get("media", {}).get("sources", [])
        if not content.get("file", "").endswith("m3u8")
    ]



from urllib.parse import unquote
import base64


def decipher(encrypted_url: str):

    s1, s2 = encrypted_url[:16], encrypted_url[16:]

    decrypted = unquote("".join(chr(x) for x in base64.b64decode(s2)))
    mapper = {byte_index: byte_index for byte_index in range(0x100)}
    xcrypto = 0

    for byte_index in range(0x100):
        xcrypto = (
            xcrypto + mapper.get(byte_index) + ord(s1[byte_index % len(s1)])
        ) % 0x100
        mapper[byte_index], mapper[xcrypto] = mapper[xcrypto], mapper[byte_index]

    xcryptoz, xcryptoy = 0, 0

    for character in decrypted:
        xcryptoy = (xcryptoy + 1) % 0x100
        xcryptoz = (xcryptoz + mapper.get(xcryptoy)) % 0x100
        mapper[xcryptoy], mapper[xcryptoz] = mapper[xcryptoz], mapper[xcryptoy]
        yield chr(
            ord(character) ^ mapper[(mapper[xcryptoy] + mapper[xcryptoz]) % 0x100]
        )



from functools import partial

import regex
import yarl

from ....config import ANIMEONSEN
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(ANIMEONSEN, extra_regex=r"/watch\?v=([^&?/]+)")

METADATA_REGEX = regex.compile(r'url: "(.+?)"')


def get_metadata(session, url):
    return session.get(url, headers={"origin": ANIMEONSEN.rstrip("/")}).json()


def stream_url_from_metadata(metadata):
    uri = metadata.get("player", {}).get("uri", {})
    return [
        {
            "stream_url": uri.get("stream"),
            "subtitle": [
                subtitle.get("location")
                for subtitle in uri.get("subtitles", {}).values()
            ],
            "headers": {
                "referer": ANIMEONSEN,
            },
        }
    ]


def fetcher(session, url, check, match):

    content_uri = yarl.URL(ANIMEONSEN + "watch")

    episode_page = session.get(
        content_uri.with_query({"v": match.group(1)}).human_repr()
    )
    metadata = get_metadata(session, METADATA_REGEX.search(episode_page.text).group(1))

    if check(1):
        yield partial(stream_url_from_metadata, metadata), 1

    for _ in range(2, int(metadata.get("metadata", {}).get("totalEpisodes", 2)) + 1):
        if check(_):
            yield partial(
                lambda episode: stream_url_from_metadata(
                    get_metadata(
                        session,
                        METADATA_REGEX.search(
                            session.get(
                                content_uri.with_query(
                                    {"v": match.group(1), "ep": episode}
                                ).human_repr()
                            ).text
                        ).group(1),
                    )
                ),
                _,
            ), _



import logging

COLORS = {
    "WARNING": "\x1b[33m",
    "INFO": "\x1b[32m",
    "DEBUG": "\x1b[36m",
    "CRITICAL": "\x1b[91m",
    "ERROR": "\x1b[31m",
}


class LoggingFormatter(logging.Formatter):

    use_color = True
    logging_format = "[\x1b[35m%(filename)s:%(lineno)d\x1b[39m - %(asctime)s - %(name)s: %(levelname)s] %(message)s"

    def __init__(self):
        super().__init__(fmt=self.logging_format)

    def format(self, record: logging.LogRecord) -> str:

        if self.use_color:
            color = COLORS.get(record.levelname)
            record.levelname = "{}{}\x1b[39m".format(color, record.levelname[0])
            if record.name:
                record.name = "{}{}\x1b[39m".format(color, record.name)

        return super().format(record)


class FileLoggingFormatter(logging.Formatter):

    logging_format = "[%(filename)s:%(lineno)d\x1b - %(asctime)s - %(name)s: %(levelname)s] %(message)s"

    def __init__(self):
        super().__init__(fmt=self.logging_format)


class Logger(logging.Logger):
    def __init__(self, name):
        super().__init__(name)
        self.propagate = False

        formatter = LoggingFormatter()

        if self.FILE_STREAM is not None:
            file_stream_handler = logging.FileHandler(
                self.FILE_STREAM, "a", encoding="utf-8"
            )
            file_stream_handler.setFormatter(FileLoggingFormatter())
            self.addHandler(file_stream_handler)
        else:
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(formatter)
            self.addHandler(stream_handler)


def configure_logger(logger_class=Logger):
    logging.setLoggerClass(logger_class)



import shutil
import subprocess
from pathlib import Path

from ...config import PLAYERS


def supported_streamers():
    for player, player_info in PLAYERS.items():
        if Path(player_info.get("executable")).exists() or shutil.which(
            player_info.get("executable")
        ):
            yield player, player_info


def start_streaming_ffplay(executable, stream_url, opts, *, headers=None, **kwargs):
    """
    ffplay does not support subtitles from a url.
    """
    args = [executable, stream_url] + (opts or [])

    if headers:
        args.extend(
            ("-headers", "\r\n".join("{}:{}".format(k, v) for k, v in headers.items()))
        )

    content_title = kwargs.pop("content_title", "")

    if content_title:
        args.extend(["-metadata", "title={}".format(content_title)])

    return subprocess.Popen(args)


def start_streaming_mpv(executable, stream_url, opts, *, headers=None, **kwargs):
    args = [executable, stream_url, "--force-window=immediate"] + (opts or [])

    if headers:
        args.append(
            "--http-header-fields=%s"
            % "\r\n".join("{}:{}".format(k, v) for k, v in headers.items())
        )

    content_title = kwargs.pop("content_title", "")
    subtitles = kwargs.pop("subtitles", []) or []

    if content_title:
        args.append("--title=%s" % content_title)

    args.extend("--sub-file={}".format(sub) for sub in subtitles)

    return subprocess.Popen(args)


def start_streaming_celluloid(executable, stream_url, opts, *, headers=None, **kwargs):
    args = [executable, stream_url, "--new-window"] + (opts or [])

    if headers:
        args.append(
            "--mpv-http-header-fields=%s"
            % "\r\n".join("{}:{}".format(k, v) for k, v in headers.items())
        )

    content_title = kwargs.pop("content_title", "")
    subtitles = kwargs.pop("subtitles", []) or []

    if content_title:
        args.append("--mpv-title=%s" % content_title)

    args.extend("--mpv-sub-file={}".format(sub) for sub in subtitles)

    return subprocess.Popen(args)


def start_streaming_vlc(executable, stream_url, opts, *, headers=None, **kwargs):
    args = [executable, stream_url] + (opts or [])

    if headers:
        if headers.get("referer"):
            args.append("--http-referrer={}".format(headers.get("referer")))

        if headers.get("user-agent"):
            args.append("--http-user-agent={}".format(headers.get("user-agent")))

    subtitles = kwargs.pop("subtitles", []) or []
    args.extend("--sub-file={}".format(sub) for sub in subtitles)

    return subprocess.Popen(args)


def start_streaming_android(executable, stream_url, opts, *, headers=None, **kwargs):

    args = [
        executable,
        "start",
        "--user",
        "0",
        "-a",
        "android.intent.action.VIEW",
        "-d",
        stream_url,
    ]

    if headers:
        args.extend(
            (
                "-e",
                "http-header-fields=%s"
                % "\r\n".join("{}:{}".format(k, v) for k, v in headers.items()),
            )
        )

    subtitles = kwargs.pop("subtitles", []) or []

    if subtitles:
        for subtitle in subtitles:
            args.extend(("-e", "sub-file={}".format(subtitle)))

    args.extend(opts or [])
    args.extend(("-n", "is.xyz.mpv/.MPVActivity"))

    return subprocess.Popen(args)


PLAYER_MAPPING = {
    "mpv": start_streaming_mpv,
    "iina": start_streaming_mpv,
    "vlc": start_streaming_vlc,
    "celluloid": start_streaming_celluloid,
    "ffplay": start_streaming_ffplay,
    "android": start_streaming_android,
}


def handle_streamer(player_opts, **kwargs):
    supported = [*supported_streamers()]
    user_selection = [
        k for k, v in kwargs.items() if v and any(k == p for p, _ in supported)
    ]

    if not user_selection:
        return False

    player = user_selection.pop(0)
    player_info = PLAYERS.get(player)

    return lambda *a, **k: start_streaming(
        player,
        player_info.get("executable"),
        opts=player_info.get("opts", []) + (player_opts or []),
        *a,
        **k
    )


def start_streaming(player, executable, stream_url, *, headers=None, **kwargs):
    return PLAYER_MAPPING.get(player, lambda *args, **kwargs: False)(
        executable, stream_url, headers=headers, **kwargs
    )



import json
from collections import defaultdict
from functools import partial

import lxml.html as htmlparser
import regex

from ....config import CRUNCHYROLL
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(CRUNCHYROLL, extra_regex=r"/([^?/&]+)")

CONTENT_METADATA = regex.compile(r"vilos\.config\.media = (\{.+\})")


def get_subtitle(subtitles, lang="enUS"):
    for sub in subtitles:
        if sub.get("language") == lang:
            yield sub.get("url")


def get_stream_urls(session, episode_data):
    for episode_page, title in episode_data:
        json_content = json.loads(
            CONTENT_METADATA.search(session.get(episode_page).text).group(1)
        )
        metadata = json_content.get("metadata")

        for stream in json_content.get("streams"):
            if (
                stream.get("format")
                in [
                    "adaptive_dash",
                    "adaptive_hls",
                    "multitrack_adaptive_hls_v2",
                    "vo_adaptive_dash",
                    "vo_adaptive_hls",
                ]
                and stream.get("hardsub_lang") in [None, "enUS"]
            ):
                yield_content = {
                    "stream_url": stream.get("url"),
                    "title": "{} ({})".format(metadata.get("title"), title)
                    if title
                    else metadata.get("title"),
                }

                if stream.get("hardsub_lang") is None:
                    yield_content.update(
                        {
                            "subtitle": [*get_subtitle(json_content.get("subtitles"))],
                            "download": False,
                        }
                    )

                yield yield_content


def group_content(slug, html_element):

    episodes = defaultdict(list)

    for element in html_element.cssselect("a.episode")[::-1]:
        episode_match = regex.search(
            "^/{}/episode-(\d+)".format(regex.escape(slug)), element.get("href")
        )
        episodes[int(episode_match.group(1)) if episode_match else 0].append(
            (CRUNCHYROLL + element.get("href", "").strip("/"), element.get("title"))
        )

    return episodes


def fetcher(session, url, check, match):

    slug = match.group(1)
    url = CRUNCHYROLL + slug

    episode_urls = sorted(
        group_content(slug, htmlparser.fromstring(session.get(url).text)).items()
    )

    if not episode_urls:
        us_catalouge_session = session.get(
            "https://raw.githubusercontent.com/justfoolingaround/animdl-provider-benchmarks/master/api/raw"
        ).text
        session.cookies.update({"session_id": us_catalouge_session})
        episode_urls = sorted(
            group_content(slug, htmlparser.fromstring(session.get(url).text)).items()
        )

    for episode_number, episode_data in episode_urls:
        if check(episode_number):
            yield partial(
                lambda e: list(get_stream_urls(session, e)), episode_data
            ), episode_number



from functools import partial

import lxml.html as htmlparser

from ....config import HAHO
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(HAHO, extra_regex=r"/anime/([^?&/]+)")


def extract_urls(session, episode_page):
    episode_page_content = session.get(episode_page)
    embed_page = (
        htmlparser.fromstring(episode_page_content.text).cssselect("iframe") or [{}]
    )[0].get("src")
    streams_page = session.get(embed_page, headers={"referer": episode_page})

    def _get_quality(quality_string):
        q = quality_string.rstrip("pP")
        if q.isdigit():
            return int(q)
        return None

    yield from (
        {"stream_url": _.get("src"), "quality": _get_quality(_.get("title"))}
        for _ in htmlparser.fromstring(streams_page.text).cssselect("source")
    )


def fetcher(session, url, check, match):
    url = match.group(0)

    episode_list_page = session.get(url)
    count = int(
        htmlparser.fromstring(episode_list_page.text)
        .cssselect("span.badge")[0]
        .text_content()
    )

    for episode in range(1, count + 1):
        if check(episode):
            yield partial(
                lambda c: list(extract_urls(session, c)), "{}/{:d}".format(url, episode)
            ), episode



from functools import partial

import lxml.html as htmlparser
import regex

from ....config import GOGOANIME
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(
    GOGOANIME, extra_regex=r"/(?:([^&?/]+)-episode-\d+|category/([^&?/]+))"
)

ANIME_RE = construct_site_based_regex(GOGOANIME, extra_regex=r"/([^&?/]+)-episode-\d+")

EPISODE_LOAD_AJAX = "https://ajax.gogo-load.com/ajax/load-list-episode"
SITE_URL = GOGOANIME


def get_episode_list(session, anime_id):
    """
    Fetch all the episodes' url from GogoAnime using.
    """
    episode_page = session.get(
        EPISODE_LOAD_AJAX,
        params={
            "ep_start": "0",
            "ep_end": "2000",
            "id": anime_id,
        },
    )
    content = htmlparser.fromstring(episode_page.text)

    for episode in content.xpath('//ul[@id="episode_related"]/li/a'):
        yield SITE_URL + episode.get("href", "").strip()


def get_anime_id(html_content):
    content = html_content.xpath('//input[@id="movie_id"]')
    assert content, "No GGA Anime ID found."
    return int(content[0].get("value", 0))


def convert_to_anime_page(url):
    match = ANIME_RE.search(url)
    if match:
        return SITE_URL + "/category/%s" % match.group(1)
    return url


def get_quality(url_text):
    match = regex.search(r"(\d+)P", url_text)
    if not match:
        return None
    return int(match.group(1))


def get_embed_page(session, episode_url):
    response = session.get(episode_url)
    content_parsed = htmlparser.fromstring(response.text)
    return "https:{}".format(content_parsed.cssselect("iframe")[0].get("src"))


def fetcher(session, url, check, match):
    url = convert_to_anime_page(url)

    anime_page = session.get(url)
    content_id = get_anime_id(htmlparser.fromstring(anime_page.text))

    episodes = reversed([*get_episode_list(session, content_id)])

    for index, episode in enumerate(episodes, 1):
        if check(index):
            yield partial(
                lambda e: [
                    {
                        "stream_url": get_embed_page(session, e),
                        "further_extraction": ("gogoplay", {}),
                    }
                ],
                episode,
            ), index



import httpx
import logging

from contextlib import suppress

headers = httpx.Headers(
    {
        b"Accept": b"*/*",
        b"Accept-Encoding": b"gzip, deflate",
        b"Connection": b"keep-alive",
        b"Referer": "https://google.com/",
        b"User-Agent": b"animdl/1.4.20",
    }
)


def get_safeoverride(f):
    def inner(*args, **kwargs):
        with suppress():
            return f(*args, **kwargs)
        return

    return inner


class AnimeHttpClient(httpx.Client):

    http_logger = logging.getLogger("animdl-http")


client = AnimeHttpClient(headers=headers, timeout=30.0)

client.__del__ = get_safeoverride(client.__del__)



import json as json_
import logging

import click

from ...config import DEFAULT_PROVIDER
from ..helpers import bannerify
from ..helpers.searcher import link
from ..http_client import client


@click.command(name="search", help="Search for an anime in the provider.")
@click.argument("query", required=True)
@click.option(
    "-p",
    "--provider",
    help="Provider to search in.",
    required=False,
    default=DEFAULT_PROVIDER,
)
@click.option(
    "-ll", "--log-level", help="Set the integer log level.", type=int, default=20
)
@click.option(
    "--log-file",
    help="Set a log file to log everything to.",
    required=False,
)
@click.option(
    "-j",
    "--json",
    help="An integer that determines where to begin the grabbing from.",
    is_flag=True,
    flag_value=True,
)
@bannerify
def animdl_search(query, json, provider, **kwargs):
    logger = logging.getLogger("searcher")
    session = client

    if provider not in link:
        logger.critical(
            "{!r} is not supported at the moment. Selecting the default {!r}.".format(
                provider, DEFAULT_PROVIDER
            )
        )
        provider = DEFAULT_PROVIDER

    genexp = link.get(provider)(session, query)

    for i, search_data in enumerate(genexp):
        logger.info(
            "[#{:02d}] {name} \x1b[33m{anime_url}\x1b[39m".format(i, **search_data)
        ) if not json else print(json_.dumps(search_data))



import regex
from base64 import b64decode

from .decrypter import decrypt

YTSM = regex.compile(r"ysmm = '([^']+)")
KWIK_F_URL = regex.compile(r"[&f]/(.+)")
LOCKED_AD = regex.compile(r"&url=([^&]+)")

KWIK_PARAMS_RE = regex.compile(r'\("(\w+)",\d+,"(\w+)",(\d+),(\d+),\d+\)')
KWIK_D_URL = regex.compile(r'action="([^"]+)"')
KWIK_D_TOKEN = regex.compile(r'value="([^"]+)"')


def decode_adfly(coded_key):
    r, j = "", ""
    for n, l in enumerate(coded_key):
        if not n % 2:
            r += l
        else:
            j = l + j

    encoded_uri = list(r + j)
    numbers = ((i, n) for i, n in enumerate(encoded_uri) if str.isdigit(n))
    for first, second in zip(numbers, numbers):
        xor = int(first[1]) ^ int(second[1])
        if xor < 10:
            encoded_uri[first[0]] = str(xor)

    return b64decode(("".join(encoded_uri)).encode("utf-8"))[16:-16].decode(
        "utf-8", errors="ignore"
    )


def bypass_adfly(session, adfly_url):
    """
    Instant extraction of the adfly url for bots.
    """
    response_code = 302
    while response_code != 200:
        adfly_content = session.get(
            session.get(adfly_url, allow_redirects=False).headers.get("location"),
            allow_redirects=False,
        )
        response_code = adfly_content.status_code
    return decode_adfly(YTSM.search(adfly_content.text).group(1))


def get_stream_url_from_kwik(session, adfly_url):
    f_content = session.get(
        bypass_adfly(session, adfly_url), headers={"referer": adfly_url}
    )
    decrypted = decrypt(*KWIK_PARAMS_RE.search(f_content.text).group(1, 2, 3, 4))

    code = 419
    while code != 302:
        content = session.post(
            KWIK_D_URL.search(decrypted).group(1),
            allow_redirects=False,
            data={"_token": KWIK_D_TOKEN.search(decrypted).group(1)},
            headers={
                "referer": str(f_content.url),
                "cookie": f_content.headers.get("set-cookie"),
            },
        )
        code = content.status_code

    return content.headers.get("location")



import regex

STREAMLARE = "https://streamlare.com/"

CONTENT_ID = regex.compile(r"/e/([^?#&/]+)")


def extract(session, url, **opts):

    content_id = CONTENT_ID.search(url).group(1)

    def fast_yield():
        for _, streams in (
            session.post(
                STREAMLARE + "api/video/get",
                headers={"x-requested-with": "XMLHttpRequest"},
                json={"id": content_id},
            )
            .json()
            .get("result")
            .items()
        ):
            yield {"stream_url": streams.get("src"), "headers": {"referer": STREAMLARE}}

    return list(fast_yield())



import functools
import logging
import traceback

import regex
from click import prompt

from ...codebase import downloader, extractors
from .fun import bannerify, choice, create_random_titles, stream_judiciary, to_stdout
from .player import *
from .searcher import link as processor_link
from .processors import get_searcher, process_query
from .intelliq import filter_quality

fe_logger = logging.getLogger("further-extraction")


def inherit_stream_meta(parent, streams, *, exempt=["headers", "stream_url"]):
    for stream in streams:
        stream.update({_: __ for _, __ in parent.items() if _ not in exempt})
        yield stream


def further_extraction(session, stream):
    extractor, options = stream.pop("further_extraction", (None, None))

    if extractor:
        for ext_module, ext in extractors.iter_extractors():
            if ext == extractor:
                try:
                    return functools.reduce(
                        lambda x, y: x + y,
                        list(
                            further_extraction(session, inherited_stream)
                            for inherited_stream in inherit_stream_meta(
                                stream,
                                ext_module.extract(session, stream.get("stream_url")),
                                **options
                            )
                        ),
                        [],
                    )
                except Exception as e:
                    fe_logger.error(
                        "Extraction from {!r} failed due to: {!r}.".format(ext, e)
                    )
    else:
        return [stream]

    return []


def ensure_extraction(session, stream_uri_caller):
    for stream in stream_uri_caller():
        if "further_extraction" in stream:
            yield from further_extraction(session, stream)
        else:
            yield stream


def get_range_conditions(range_string):
    for matches in regex.finditer(r"(?:([0-9]*)[:\-.]([0-9]*)|([0-9]+))", range_string):
        start, end, singular = matches.groups()
        if start and end and int(start) > int(end):
            start, end = end, start
        yield (lambda x, s=singular: int(s) == x) if singular else (
            lambda x: True
        ) if not (start or end) else (
            lambda x, s=start: x >= int(s)
        ) if start and not end else (
            lambda x, e=end: x <= int(e)
        ) if not start and end else (
            lambda x, s=start, e=end: int(s) <= x <= int(e)
        )


def get_check(range_string):
    if not range_string:
        return lambda *args, **kwargs: True
    return lambda x: any(
        condition(x) for condition in get_range_conditions(range_string)
    )


def ask(log_level, **prompt_kwargs):

    if log_level > 20:
        return prompt_kwargs.get("default")

    return prompt(**prompt_kwargs)


DOWNLOAD_ERROR_MESSAGE = """
\x1b[34mDownload failure #{0:02d}\x1b[39m

Download Arguments: {1}

Raw exception: {2!r}

Complete traceback: {3}\
"""


def download(
    session, logger, content_dir, outfile_name, stream_urls, quality, **kwargs
):
    downloadable_content = filter_quality(stream_urls, quality)

    if not downloadable_content:
        return False, "No downloadable content found."

    errors = []

    for download_data in iter(downloadable_content):
        if "further_extraction" in download_data:
            try:
                further_status, further_yield = download(
                    session,
                    logger,
                    content_dir,
                    outfile_name,
                    further_extraction(session, download_data),
                    quality,
                    **kwargs
                )
                if further_status:
                    return further_status, further_yield
                continue
            except Exception as e:
                logger.critical(
                    "Could not extract from the stream due to {!r}. falling back to other streams.".format(
                        e
                    )
                )
                continue

        content_url = download_data.get("stream_url")
        content_headers = download_data.get("headers")

        try:
            return True, downloader.handle_download(
                session,
                content_url,
                content_headers,
                content_dir,
                outfile_name
                if not download_data.get("is_torrent", False)
                else download_data.get("torrent_name"),
                preferred_quality=quality,
                subtitles=download_data.get("subtitle", []),
                **kwargs
            )
        except Exception as e:
            logger.critical(
                "Oops, due to {!r}, this stream has been rendered unable to download.".format(
                    e
                )
            )

            errors.append((download_data, e, traceback.format_exc()))

    logger.critical(
        "No downloads were done. Use log level DEBUG or, -ll 0 to get complete tracebacks for the failed downloads."
    )

    for count, (download_data, exception, tb) in enumerate(errors, 1):
        logger.debug(DOWNLOAD_ERROR_MESSAGE.format(count, download_data, exception, tb))

    return False, "All stream links were undownloadable."



import regex


def bypass_ddos_guard(session, base_uri):
    js_bypass_uri = regex.search(
        r"'(.*?)'", session.get("https://check.ddos-guard.net/check.js").text
    ).group(1)

    session.get(base_uri + js_bypass_uri)



from functools import partial

import lxml.html as htmlparser

from ....config import NYAASI
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(NYAASI, extra_regex=r"/view/([^?&/]+)")


def fetcher(session, url, check):

    html_element = htmlparser.fromstring(session.get(url).text)
    content_title = html_element.cssselect("h3.panel-title")[0].text_content()

    for n, magnets in enumerate(html_element.cssselect('a[href^="magnet:?"]'), 1):
        yield partial(
            lambda x: [
                {
                    "stream_url": x,
                    "is_torrent": True,
                    "torrent_name": content_title.strip(),
                }
            ],
            magnets.get("href"),
        ), n



import logging

import click

from ...codebase import providers
from ..helpers import bannerify, ensure_extraction
from ..http_client import client


@click.command(name="test", help="Test the scrapability power.")
@click.option(
    "-x",
    help="A list of certain sites (full anime url) to be explicity used for testing.",
    default=[],
    required=False,
    multiple=True,
)
@click.option(
    "-e",
    help="Episode number to bestow the testing upon",
    default=1,
    required=False,
    type=int,
)
@click.option(
    "--log-file",
    help="Set a log file to log everything to.",
    required=False,
)
@click.option(
    "-ll", "--log-level", help="Set the integer log level.", type=int, default=20
)
@bannerify
def animdl_test(x, e, **kwargs):
    session = client
    SITE_LIST = {
        "9anime": "https://9anime.to/watch/one-piece.ov8",
        "crunchyroll": "https://www.crunchyroll.com/one-piece",
        "allanime": "https://allanime.site/anime/ReooPAxPMsHM4KPMY",
        "animeout": "https://www.animeout.xyz/download-one-piece-episodes-latest/",
        "animixplay": "https://animixplay.to/v1/one-piece",
        "animtime": "https://animtime.com/title/5",
        "gogoanime": "https://gogoanime.cm/category/one-piece",
        "tenshi": "https://tenshi.moe/anime/kjfrhu3s",
        "twist": "https://twist.moe/a/one-piece",
    }

    if not x:
        x = SITE_LIST.values()

    logger = logging.getLogger("tests")

    for site in x:
        logger.info("Attempting to scrape anime from {!r}.".format(site))
        try:
            links = [*providers.get_appropriate(session, site, lambda r: r == e)]
            if not links:
                raise Exception("No stream urls found on {!r}.".format(site))
            for link_cb, _ in links:
                for stream in ensure_extraction(session, link_cb):
                    print("\t - \x1b[32m{stream_url}\x1b[39m".format_map(stream))
            logger.info(
                "Scraping from {!r} was \x1b[32msuccessful\x1b[39m.".format(site)
            )
        except Exception as excep:
            logger.error(
                "\x1b[31mFailed to scrape any urls from {!r} due to {!r}.\x1b[39m".format(
                    site, excep
                )
            )



from setuptools import setup, find_packages
from animdl.core.__version__ import __core__

setup(
    name="animdl",
    version=__core__,
    author="kr@justfoolingaround",
    author_email="kr.justfoolingaround@gmail.com",
    description="A highly efficient, fast, powerful and light-weight anime downloader and streamer for your favorite anime.",
    packages=find_packages(),
    url="https://github.com/justfoolingaround/animdl",
    keywords=[
        "stream",
        "anime",
        "download",
        "anime-downloader",
        "twist",
        "9anime",
        "gogoanime",
        "animepahe",
        "4anime",
        "anime-streamer",
        "fouranime",
        "animixplay",
    ],
    install_requires=[
        "anitopy",
        "click",
        "comtypes",
        "cssselect",
        "httpx==0.19.0",
        "requests",
        "pyyaml",
        "yarl",
        "lxml",
        "tqdm",
        "regex==2021.10.8",
        "pycryptodomex",
    ],
    entry_points="""
        [console_scripts]
        animdl=animdl.__main__:__animdl_cli__
    """,
)



import regex
from collections import defaultdict
from functools import partial

import lxml.html as htmlparser

REGEX = regex.compile(
    r"(?:https?://)?(?:\S+\.)?(?P<host>domdom\.stream|bestanime3\.xyz|kawaiifu\.com)/(?P<episode_page>anime/)?(?P<type>season/[^/]+|.+)/(?P<slug>[^?&#]+)"
)


def get_int(content):
    d = regex.search(r"[0-9]+", content)
    if d:
        return int(d.group(0))


def extract_stream_urls(session, urls):
    for url in urls:
        html_element = htmlparser.fromstring(session.get(url).text)
        for source in html_element.cssselect("source"):
            yield {
                "quality": get_int(source.get("data-quality")),
                "stream_url": source.get("src"),
                "headers": {"referer": url},
            }


def get_from_url(session, url):
    episodes = defaultdict(list)
    html_element = htmlparser.fromstring(session.get(url).text)

    for servers in html_element.cssselect(".list-server"):
        for element in servers.cssselect(".list-ep a"):
            episodes[get_int(element.text_content()) or 0].append(element.get("href"))
    return episodes


def fetcher(session, url, check, match):

    if match.group("host") == "kawaiifu.com":
        url += "-episode"

    for episode, episode_urls in sorted(
        get_from_url(session, url).items(), key=lambda x: x[0]
    ):
        if check(episode):
            yield partial(
                lambda s, x: [*extract_stream_urls(s, x)], session, episode_urls
            ), episode



"""
Torrent client!
"""

import time

import httpx
import regex
from tqdm import tqdm

MAGNET_URI_REGEX = regex.compile(r"btih:([^&?]+)")


def is_supported(session, endpoint):
    try:
        session.get(endpoint + "/api/v2")
    except (httpx.ConnectError, httpx.ConnectTimeout):
        return False
    return True


def authenticate(session, endpoint, login_data):
    return session.post(endpoint + "/api/v2/auth/login", data=login_data).text == "Ok."


def wrap_with_tqdm(session, btih, endpoint, torrent_name, log_level):
    """
    Do not use this in an unauthenticated scenario.
    """

    uri = endpoint + "/api/v2/torrents/properties?hash={}".format(btih)
    torrent = session.get(uri).json()

    progress_bar = tqdm(
        disable=log_level > 20,
        total=torrent.get("total_size", 0),
        unit="B",
        unit_divisor=1024,
        unit_scale=True,
        desc="qBittorrent / {}".format(torrent_name),
        initial=torrent.get("total_downloaded", 0),
    )
    current = torrent.get("total_downloaded", 0)

    while torrent.get("eta") > 0:
        torrent = session.get(uri).json()
        progress_bar.update(torrent.get("total_downloaded", 0) - current)
        current = torrent.get("total_downloaded")

        progress_bar.total = progress_bar.total or torrent.get("total_size", 0)

        time.sleep(1)


def download_torrent(
    _, magnet_uri, content_dir, torrent_name, endpoint, login_data, *, log_level=20
):
    """
    Enqueues the torrent to the qBittorrent client. If already there, doesn't but shows the status of the torrent.
    """
    session = httpx.Client()

    auth = authenticate(session, endpoint, login_data)

    if not auth:
        raise Exception("Invalid credentials for qBittorrent server.")

    magnet_match = MAGNET_URI_REGEX.search(magnet_uri)
    if not magnet_match:
        raise Exception("Not a valid magnet url.")

    btih = magnet_match.group(1)

    existent = session.post(
        endpoint + "/api/v2/torrents/properties", params={"hash": btih}
    )

    if existent.status_code in [404, 400]:
        status = session.post(
            endpoint + "/api/v2/torrents/add",
            data={"urls": magnet_uri, "savepath": content_dir.resolve().as_posix()},
        ).text

        if status == "Fails.":
            raise Exception(
                "Could not add to the client: ({!r}, {!r})".format(
                    magnet_uri, content_dir
                )
            )

    return wrap_with_tqdm(session, btih, endpoint, torrent_name, log_level)
